/*
 * Othello
 *
 * Copyright (C) Evan Harris, 1994
 *
 * Permission is granted to freely redistribute and modify this code,
 * providing the author(s) get credit for having written it.
 */

#if !defined(CLOCKS_PER_SECOND) && defined(CLOCKS_PER_SEC)
#define CLOCKS_PER_SECOND CLOCKS_PER_SEC
#endif

#define NEWGAME -1
#define QUIT -2
#define UNKNOWN -8
#define ILLEGAL -9

void NewGame(void);
void InitDisplay(int reverse);
void EndDisplay(void);
void PutBoardStart(void);
void PutBoardEnd(void);
void PutSquare(int x, int y, int entry);
void ShowWin(int who, int winscore, int losescore);
void ShowDraw(int score);
void ThinkingOn(void);
void ThinkingOff(void);
void ShowTime(double t);
int GetMove(void);
void Poll(void);
