/*
 * GamesPak
 *
 * Copyright (C) Evan Harris, 1994
 *
 * Permission is granted to freely redistribute and modify this code,
 * providing the author(s) get credit for having written it.
 */

#define KEY_CURSORUP    -10
#define KEY_CURSORDOWN  -11
#define KEY_CURSORLEFT  -12
#define KEY_CURSORRIGHT -13

int key_getkey();
