/*
 * GamesPak
 *
 * Copyright (C) Evan Harris, 1994, 1995.
 *
 * Permission is granted to freely redistribute and modify this code,
 * providing the author(s) get credit for having written it.
 */

void RenderMousePointer(int x, int y, int light, int dark,
			int width, int height);
void RestoreUnderMousePointer(int x, int y, int width, int height,
			      int *colour);
void SaveUnderMousePointer(int x, int y, int width, int height, int *colour);
